"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.detectTestingLibraryUtils = void 0;
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const node_utils_1 = require("../node-utils");
const utils_1 = require("../utils");
const SETTING_OPTION_OFF = 'off';
const USER_EVENT_PACKAGE = '@testing-library/user-event';
const FIRE_EVENT_NAME = 'fireEvent';
const USER_EVENT_NAME = 'userEvent';
const RENDER_NAME = 'render';
function detectTestingLibraryUtils(ruleCreate, { skipRuleReportingCheck = false } = {}) {
    return (context, optionsWithDefault) => {
        let importedTestingLibraryNode = null;
        let importedCustomModuleNode = null;
        let importedUserEventLibraryNode = null;
        const customModuleSetting = context.settings['testing-library/utils-module'];
        const customRendersSetting = context.settings['testing-library/custom-renders'];
        const customQueriesSetting = context.settings['testing-library/custom-queries'];
        function isTestingLibraryUtil(node, isUtilCallback) {
            if (!node) {
                return false;
            }
            const referenceNode = node_utils_1.getReferenceNode(node);
            const referenceNodeIdentifier = node_utils_1.getPropertyIdentifierNode(referenceNode);
            if (!referenceNodeIdentifier) {
                return false;
            }
            const importedUtilSpecifier = getImportedUtilSpecifier(referenceNodeIdentifier);
            const originalNodeName = node_utils_1.isImportSpecifier(importedUtilSpecifier) &&
                importedUtilSpecifier.local.name !== importedUtilSpecifier.imported.name
                ? importedUtilSpecifier.imported.name
                : undefined;
            if (!isUtilCallback(node.name, originalNodeName)) {
                return false;
            }
            if (isAggressiveModuleReportingEnabled()) {
                return true;
            }
            return isNodeComingFromTestingLibrary(referenceNodeIdentifier);
        }
        const isAggressiveModuleReportingEnabled = () => !customModuleSetting;
        const isAggressiveRenderReportingEnabled = () => {
            const isSwitchedOff = customRendersSetting === SETTING_OPTION_OFF;
            const hasCustomOptions = Array.isArray(customRendersSetting) && customRendersSetting.length > 0;
            return !isSwitchedOff && !hasCustomOptions;
        };
        const isAggressiveQueryReportingEnabled = () => {
            const isSwitchedOff = customQueriesSetting === SETTING_OPTION_OFF;
            const hasCustomOptions = Array.isArray(customQueriesSetting) && customQueriesSetting.length > 0;
            return !isSwitchedOff && !hasCustomOptions;
        };
        const getCustomModule = () => {
            if (!isAggressiveModuleReportingEnabled() &&
                customModuleSetting !== SETTING_OPTION_OFF) {
                return customModuleSetting;
            }
            return undefined;
        };
        const getCustomRenders = () => {
            if (!isAggressiveRenderReportingEnabled() &&
                customRendersSetting !== SETTING_OPTION_OFF) {
                return customRendersSetting;
            }
            return [];
        };
        const getCustomQueries = () => {
            if (!isAggressiveQueryReportingEnabled() &&
                customQueriesSetting !== SETTING_OPTION_OFF) {
                return customQueriesSetting;
            }
            return [];
        };
        const getTestingLibraryImportNode = () => {
            return importedTestingLibraryNode;
        };
        const getCustomModuleImportNode = () => {
            return importedCustomModuleNode;
        };
        const getTestingLibraryImportName = () => {
            return node_utils_1.getImportModuleName(importedTestingLibraryNode);
        };
        const getCustomModuleImportName = () => {
            return node_utils_1.getImportModuleName(importedCustomModuleNode);
        };
        const isTestingLibraryImported = (isStrict = false) => {
            const isSomeModuleImported = !!importedTestingLibraryNode || !!importedCustomModuleNode;
            return ((!isStrict && isAggressiveModuleReportingEnabled()) ||
                isSomeModuleImported);
        };
        const isQuery = (node) => {
            const hasQueryPattern = /^(get|query|find)(All)?By.+$/.test(node.name);
            if (!hasQueryPattern) {
                return false;
            }
            if (isAggressiveQueryReportingEnabled()) {
                return true;
            }
            const customQueries = getCustomQueries();
            const isBuiltInQuery = utils_1.ALL_QUERIES_COMBINATIONS.includes(node.name);
            const isReportableCustomQuery = customQueries.some((pattern) => new RegExp(pattern).test(node.name));
            return isBuiltInQuery || isReportableCustomQuery;
        };
        const isGetQueryVariant = (node) => {
            return isQuery(node) && node.name.startsWith('get');
        };
        const isQueryQueryVariant = (node) => {
            return isQuery(node) && node.name.startsWith('query');
        };
        const isFindQueryVariant = (node) => {
            return isQuery(node) && node.name.startsWith('find');
        };
        const isSyncQuery = (node) => {
            return isGetQueryVariant(node) || isQueryQueryVariant(node);
        };
        const isAsyncQuery = (node) => {
            return isFindQueryVariant(node);
        };
        const isCustomQuery = (node) => {
            return isQuery(node) && !utils_1.ALL_QUERIES_COMBINATIONS.includes(node.name);
        };
        const isBuiltInQuery = (node) => {
            return isQuery(node) && utils_1.ALL_QUERIES_COMBINATIONS.includes(node.name);
        };
        const isAsyncUtil = (node, validNames = utils_1.ASYNC_UTILS) => {
            return isTestingLibraryUtil(node, (identifierNodeName, originalNodeName) => {
                return (validNames.includes(identifierNodeName) ||
                    (!!originalNodeName &&
                        validNames.includes(originalNodeName)));
            });
        };
        const isFireEventUtil = (node) => {
            return isTestingLibraryUtil(node, (identifierNodeName, originalNodeName) => {
                return [identifierNodeName, originalNodeName].includes('fireEvent');
            });
        };
        const isUserEventUtil = (node) => {
            const userEvent = findImportedUserEventSpecifier();
            let userEventName;
            if (userEvent) {
                userEventName = userEvent.name;
            }
            else if (isAggressiveModuleReportingEnabled()) {
                userEventName = USER_EVENT_NAME;
            }
            if (!userEventName) {
                return false;
            }
            return node.name === userEventName;
        };
        const isFireEventMethod = (node) => {
            const fireEventUtil = findImportedUtilSpecifier(FIRE_EVENT_NAME);
            let fireEventUtilName;
            if (fireEventUtil) {
                fireEventUtilName = experimental_utils_1.ASTUtils.isIdentifier(fireEventUtil)
                    ? fireEventUtil.name
                    : fireEventUtil.local.name;
            }
            else if (isAggressiveModuleReportingEnabled()) {
                fireEventUtilName = FIRE_EVENT_NAME;
            }
            if (!fireEventUtilName) {
                return false;
            }
            const parentMemberExpression = node.parent && node_utils_1.isMemberExpression(node.parent)
                ? node.parent
                : undefined;
            if (!parentMemberExpression) {
                return false;
            }
            if ([fireEventUtilName, FIRE_EVENT_NAME].includes(node.name) ||
                (experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object) &&
                    parentMemberExpression.object.name === node.name)) {
                return false;
            }
            const regularCall = experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object) &&
                parentMemberExpression.object.name === fireEventUtilName;
            const wildcardCall = node_utils_1.isMemberExpression(parentMemberExpression.object) &&
                experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object.object) &&
                parentMemberExpression.object.object.name === fireEventUtilName &&
                experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object.property) &&
                parentMemberExpression.object.property.name === FIRE_EVENT_NAME;
            return regularCall || wildcardCall;
        };
        const isUserEventMethod = (node) => {
            const userEvent = findImportedUserEventSpecifier();
            let userEventName;
            if (userEvent) {
                userEventName = userEvent.name;
            }
            else if (isAggressiveModuleReportingEnabled()) {
                userEventName = USER_EVENT_NAME;
            }
            if (!userEventName) {
                return false;
            }
            const parentMemberExpression = node.parent && node_utils_1.isMemberExpression(node.parent)
                ? node.parent
                : undefined;
            if (!parentMemberExpression) {
                return false;
            }
            if ([userEventName, USER_EVENT_NAME].includes(node.name) ||
                (experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object) &&
                    parentMemberExpression.object.name === node.name)) {
                return false;
            }
            return (experimental_utils_1.ASTUtils.isIdentifier(parentMemberExpression.object) &&
                parentMemberExpression.object.name === userEventName);
        };
        const isRenderUtil = (node) => isTestingLibraryUtil(node, (identifierNodeName, originalNodeName) => {
            if (isAggressiveRenderReportingEnabled()) {
                return identifierNodeName.toLowerCase().includes(RENDER_NAME);
            }
            return [RENDER_NAME, ...getCustomRenders()].some((validRenderName) => validRenderName === identifierNodeName ||
                (Boolean(originalNodeName) && validRenderName === originalNodeName));
        });
        const isRenderVariableDeclarator = (node) => {
            if (!node.init) {
                return false;
            }
            const initIdentifierNode = node_utils_1.getDeepestIdentifierNode(node.init);
            if (!initIdentifierNode) {
                return false;
            }
            return isRenderUtil(initIdentifierNode);
        };
        const isDebugUtil = (identifierNode) => {
            const isBuiltInConsole = node_utils_1.isMemberExpression(identifierNode.parent) &&
                experimental_utils_1.ASTUtils.isIdentifier(identifierNode.parent.object) &&
                identifierNode.parent.object.name === 'console';
            return (!isBuiltInConsole &&
                isTestingLibraryUtil(identifierNode, (identifierNodeName, originalNodeName) => {
                    return [identifierNodeName, originalNodeName]
                        .filter(Boolean)
                        .includes('debug');
                }));
        };
        const isPresenceAssert = (node) => {
            const { matcher, isNegated } = node_utils_1.getAssertNodeInfo(node);
            if (!matcher) {
                return false;
            }
            return isNegated
                ? utils_1.ABSENCE_MATCHERS.includes(matcher)
                : utils_1.PRESENCE_MATCHERS.includes(matcher);
        };
        const isAbsenceAssert = (node) => {
            const { matcher, isNegated } = node_utils_1.getAssertNodeInfo(node);
            if (!matcher) {
                return false;
            }
            return isNegated
                ? utils_1.PRESENCE_MATCHERS.includes(matcher)
                : utils_1.ABSENCE_MATCHERS.includes(matcher);
        };
        const findImportedUtilSpecifier = (specifierName) => {
            var _a;
            const node = (_a = getCustomModuleImportNode()) !== null && _a !== void 0 ? _a : getTestingLibraryImportNode();
            if (!node) {
                return undefined;
            }
            if (node_utils_1.isImportDeclaration(node)) {
                const namedExport = node.specifiers.find((n) => {
                    return (node_utils_1.isImportSpecifier(n) &&
                        [n.imported.name, n.local.name].includes(specifierName));
                });
                if (namedExport) {
                    return namedExport;
                }
                return node.specifiers.find((n) => node_utils_1.isImportNamespaceSpecifier(n));
            }
            else {
                if (!experimental_utils_1.ASTUtils.isVariableDeclarator(node.parent)) {
                    return undefined;
                }
                const requireNode = node.parent;
                if (experimental_utils_1.ASTUtils.isIdentifier(requireNode.id)) {
                    return requireNode.id;
                }
                if (!node_utils_1.isObjectPattern(requireNode.id)) {
                    return undefined;
                }
                const property = requireNode.id.properties.find((n) => node_utils_1.isProperty(n) &&
                    experimental_utils_1.ASTUtils.isIdentifier(n.key) &&
                    n.key.name === specifierName);
                if (!property) {
                    return undefined;
                }
                return property.key;
            }
        };
        const findImportedUserEventSpecifier = () => {
            if (!importedUserEventLibraryNode) {
                return null;
            }
            if (node_utils_1.isImportDeclaration(importedUserEventLibraryNode)) {
                const userEventIdentifier = importedUserEventLibraryNode.specifiers.find((specifier) => node_utils_1.isImportDefaultSpecifier(specifier));
                if (userEventIdentifier) {
                    return userEventIdentifier.local;
                }
            }
            else {
                if (!experimental_utils_1.ASTUtils.isVariableDeclarator(importedUserEventLibraryNode.parent)) {
                    return null;
                }
                const requireNode = importedUserEventLibraryNode.parent;
                if (!experimental_utils_1.ASTUtils.isIdentifier(requireNode.id)) {
                    return null;
                }
                return requireNode.id;
            }
            return null;
        };
        const getImportedUtilSpecifier = (node) => {
            var _a;
            const identifierName = (_a = node_utils_1.getPropertyIdentifierNode(node)) === null || _a === void 0 ? void 0 : _a.name;
            if (!identifierName) {
                return undefined;
            }
            return findImportedUtilSpecifier(identifierName);
        };
        const canReportErrors = () => {
            return skipRuleReportingCheck || isTestingLibraryImported();
        };
        const isNodeComingFromTestingLibrary = (node) => {
            var _a;
            const importNode = getImportedUtilSpecifier(node);
            if (!importNode) {
                return false;
            }
            const identifierName = (_a = node_utils_1.getPropertyIdentifierNode(node)) === null || _a === void 0 ? void 0 : _a.name;
            if (!identifierName) {
                return false;
            }
            return node_utils_1.hasImportMatch(importNode, identifierName);
        };
        const helpers = {
            getTestingLibraryImportNode,
            getCustomModuleImportNode,
            getTestingLibraryImportName,
            getCustomModuleImportName,
            isTestingLibraryImported,
            isGetQueryVariant,
            isQueryQueryVariant,
            isFindQueryVariant,
            isSyncQuery,
            isAsyncQuery,
            isQuery,
            isCustomQuery,
            isBuiltInQuery,
            isAsyncUtil,
            isFireEventUtil,
            isUserEventUtil,
            isFireEventMethod,
            isUserEventMethod,
            isRenderUtil,
            isRenderVariableDeclarator,
            isDebugUtil,
            isPresenceAssert,
            isAbsenceAssert,
            canReportErrors,
            findImportedUtilSpecifier,
            isNodeComingFromTestingLibrary,
        };
        const detectionInstructions = {
            ImportDeclaration(node) {
                if (!importedTestingLibraryNode &&
                    /testing-library/g.test(node.source.value)) {
                    importedTestingLibraryNode = node;
                }
                const customModule = getCustomModule();
                if (customModule &&
                    !importedCustomModuleNode &&
                    String(node.source.value).endsWith(customModule)) {
                    importedCustomModuleNode = node;
                }
                if (!importedUserEventLibraryNode &&
                    String(node.source.value) === USER_EVENT_PACKAGE) {
                    importedUserEventLibraryNode = node;
                }
            },
            [`CallExpression > Identifier[name="require"]`](node) {
                const callExpression = node.parent;
                const { arguments: args } = callExpression;
                if (!importedTestingLibraryNode &&
                    args.some((arg) => node_utils_1.isLiteral(arg) &&
                        typeof arg.value === 'string' &&
                        /testing-library/g.test(arg.value))) {
                    importedTestingLibraryNode = callExpression;
                }
                const customModule = getCustomModule();
                if (!importedCustomModuleNode &&
                    args.some((arg) => customModule &&
                        node_utils_1.isLiteral(arg) &&
                        typeof arg.value === 'string' &&
                        arg.value.endsWith(customModule))) {
                    importedCustomModuleNode = callExpression;
                }
                if (!importedCustomModuleNode &&
                    args.some((arg) => node_utils_1.isLiteral(arg) &&
                        typeof arg.value === 'string' &&
                        arg.value === USER_EVENT_PACKAGE)) {
                    importedUserEventLibraryNode = callExpression;
                }
            },
        };
        const ruleInstructions = ruleCreate(context, optionsWithDefault, helpers);
        const enhancedRuleInstructions = {};
        const allKeys = new Set(Object.keys(detectionInstructions).concat(Object.keys(ruleInstructions)));
        allKeys.forEach((instruction) => {
            enhancedRuleInstructions[instruction] = (node) => {
                var _a, _b;
                if (instruction in detectionInstructions) {
                    (_a = detectionInstructions[instruction]) === null || _a === void 0 ? void 0 : _a.call(detectionInstructions, node);
                }
                if (canReportErrors() && ruleInstructions[instruction]) {
                    return (_b = ruleInstructions[instruction]) === null || _b === void 0 ? void 0 : _b.call(ruleInstructions, node);
                }
            };
        });
        return enhancedRuleInstructions;
    };
}
exports.detectTestingLibraryUtils = detectTestingLibraryUtils;
